# flake8: noqa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions

STOWAWAY_LABELS = {
    "gefyra.dev/app": "stowaway",
    "gefyra.dev/role": "connection",
    "gefyra.dev/provider": "stowaway",
}


def data(params: "GefyraInstallOptions") -> list[dict]:
    stowaway_labels = STOWAWAY_LABELS.copy()
    stowaway_annotations = {}
    if params.service_labels:
        try:
            stowaway_labels.update(params.service_labels)
        except IndexError:
            raise ValueError(
                f"Invalid service-labels format. Please use the form key=value."
            )
    if params.service_annotations:
        try:
            for annotation in params.service_annotations:
                try:
                    # handle cli params as key=value
                    key, value = annotation[0].split("=")
                    stowaway_annotations[key] = value
                except ValueError:
                    # handle preset values
                    stowaway_annotations.update(params.service_annotations)
        except IndexError:
            raise ValueError(
                f"Invalid service-annotations format. Please use the form key=value."
            )
    if params.service_type.lower() == "nodeport":
        type = "NodePort"
        ports = [
            {
                "name": "gefyra-wireguard",
                "port": 51820,
                "targetPort": 51820,
                "nodePort": params.service_port,
                "protocol": "UDP",
            }
        ]
    elif params.service_type.lower() == "loadbalancer":
        type = "LoadBalancer"
        ports = [
            {
                "name": "gefyra-wireguard",
                "port": params.service_port,
                "targetPort": 51820,
                "protocol": "UDP",
            }
        ]
    else:
        raise ValueError(f"Unknown service type: {params.service_type}")
    return [
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {
                "name": "gefyra-stowaway-wireguard",
                "namespace": params.namespace,
                "labels": stowaway_labels,
                "annotations": stowaway_annotations,
            },
            "spec": {
                "type": type,
                "ports": ports,
                "selector": STOWAWAY_LABELS,
            },
        }
    ]



import logging

from gefyra.api import stopwatch
from gefyra.configuration import ClientConfiguration
from gefyra.exceptions import ClientConfigurationError
from gefyra.types import (
    GefyraClientStatus,
    GefyraClusterStatus,
    GefyraStatus,
    StatusSummary,
)


logger = logging.getLogger(__name__)


def _get_client_status(config: ClientConfiguration) -> GefyraClientStatus:
    from docker.errors import NotFound
    from gefyra.local import CARGO_ENDPOINT_LABEL, VERSION_LABEL

    # these are the default values
    _status = GefyraClientStatus(
        cargo=False,
        version="",
        cargo_image=config.CARGO_IMAGE,
        network=False,
        connection=False,
        containers=0,
        bridges=0,
        kubeconfig=config.KUBE_CONFIG_FILE,
        context=config.KUBE_CONTEXT,
        cargo_endpoint="",
    )
    try:
        logger.debug("Checking cargo container running")
        cargo_container = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)
        if cargo_container.status == "running":
            _status.cargo = True
            _status.cargo_endpoint = cargo_container.attrs["Config"]["Labels"].get(
                CARGO_ENDPOINT_LABEL
            )
            _status.version = cargo_container.attrs["Config"]["Labels"].get(
                VERSION_LABEL
            )
            _status.cargo_image = cargo_container.image.tags[0]
    except NotFound:
        pass
    try:
        logger.debug("Checking gefyra network available")
        gefyra_net = config.DOCKER.networks.get(f"{config.NETWORK_NAME}")
        _status.network = True
        _status.containers = (
            len(gefyra_net.containers) - 1
            if _status.cargo is True
            else len(gefyra_net.containers)
        )
    except NotFound:
        return _status

    from gefyra.local.cargo import probe_wireguard_connection

    try:
        logger.debug("Probing wireguard connection")
        if _status.cargo:
            probe_wireguard_connection(config)
            _status.connection = True
    except RuntimeError:
        return _status

    from gefyra.local.bridge import get_all_gefyrabridges

    logger.debug("Counting all active bridges")
    _status.bridges = len(get_all_gefyrabridges(config))
    return _status


def _get_cluster_status(config: ClientConfiguration) -> GefyraClusterStatus:
    from kubernetes.client import ApiException, V1Pod
    from kubernetes.config import ConfigException
    from urllib3.exceptions import MaxRetryError

    # these are the default values
    _status = GefyraClusterStatus(
        connected=False,
        operator=False,
        operator_image="",
        stowaway=False,
        stowaway_image="",
        namespace=False,
        operator_webhook=False,
    )
    # check if connected to the cluster
    try:
        logger.debug("Reading API resources from Kubernetes")
        config.K8S_CORE_API.get_api_resources(_request_timeout=(1, 5))
        _status.connected = True
    except (ApiException, ConfigException, MaxRetryError):
        return _status
    # check if gefyra namespace is available
    try:
        logger.debug("Reading gefyra namespace")
        config.K8S_CORE_API.read_namespace(
            name=config.NAMESPACE, _request_timeout=(1, 5)
        )
        _status.namespace = True
    except ApiException:
        return _status
    # check if the Gefyra operator is running and ready
    try:
        logger.debug("Checking operator deployment")
        operator_deploy = config.K8S_APP_API.read_namespaced_deployment(
            name="gefyra-operator", namespace=config.NAMESPACE, _request_timeout=(1, 5)
        )
        if (
            operator_deploy.status.ready_replicas
            and operator_deploy.status.ready_replicas >= 1
        ):
            _status.operator = True
            _status.operator_webhook = True
            _status.operator_image = operator_deploy.spec.template.spec.containers[
                0
            ].image
    except ApiException:
        return _status

    # check if the Gefyra operator is running and ready
    try:
        logger.debug("Checking Stowaway endpoint")
        stowaway_pod: V1Pod = config.K8S_CORE_API.read_namespaced_pod(
            name="gefyra-stowaway-0",
            namespace=config.NAMESPACE,
            _request_timeout=(1, 5),
        )
        if stowaway_pod.status.container_statuses[0].ready:
            _status.stowaway = True
            _status.stowaway_image = stowaway_pod.spec.containers[0].image
    except ApiException as e:
        logger.warning(e)
        pass

    return _status


@stopwatch
def status(connection_name: str = "") -> GefyraStatus:
    import urllib3

    # Check if kubeconfig is available through running Cargo
    config = ClientConfiguration(connection_name=connection_name)

    cluster = _get_cluster_status(config)
    try:
        client = _get_client_status(config)
    except urllib3.exceptions.MaxRetryError as e:
        raise ClientConfigurationError(
            f"Cannot reach cluster on {e.pool.host}:{e.pool.port}\n"
            f"Used kubeconfig: {config.KUBE_CONFIG_FILE} from Gefyra connection {connection_name}\n"
            f"Please delete the connection and create a new one. Use: gefyra connection remove {connection_name}"
        )

    if client.connection:
        summary = StatusSummary.UP
    else:
        if client.cargo or (
            cluster.connected
            and cluster.operator
            and cluster.stowaway
            and cluster.operator_webhook
        ):
            summary = StatusSummary.INCOMPLETE
        else:
            summary = StatusSummary.DOWN
    return GefyraStatus(cluster=cluster, client=client, summary=summary)



# flake8: noqa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions


def data(params: "GefyraInstallOptions") -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "Service",
            "metadata": {"name": "gefyra-admission", "namespace": params.namespace},
            "spec": {
                "selector": {
                    "gefyra.dev/app": "gefyra-operator",
                    "gefyra.dev/role": "webhook",
                },
                "ports": [{"protocol": "TCP", "port": 443, "targetPort": 9443}],
            },
        },
        {
            "apiVersion": "admissionregistration.k8s.io/v1",
            "kind": "ValidatingWebhookConfiguration",
            "metadata": {"name": "gefyra.dev"},
            "webhooks": [
                {
                    "admissionReviewVersions": ["v1", "v1beta1"],
                    "clientConfig": {
                        "caBundle": "LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0tCk1JSUZJekNDQXd1Z0F3SUJBZ0lVVUpxcE1PSkdH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\nIENFUlRJRklDQVRFLS0tLS0K\n",
                        "service": {
                            "name": "gefyra-admission",
                            "namespace": params.namespace,
                            "path": "/client-parameters",
                        },
                    },
                    "failurePolicy": "Fail",
                    "matchPolicy": "Equivalent",
                    "name": "client-parameters.gefyra.dev",
                    "namespaceSelector": {},
                    "objectSelector": {},
                    "rules": [
                        {
                            "apiGroups": ["gefyra.dev"],
                            "apiVersions": ["v1"],
                            "operations": ["CREATE", "UPDATE"],
                            "resources": ["gefyraclients"],
                            "scope": "*",
                        }
                    ],
                    "sideEffects": "None",
                    "timeoutSeconds": 30,
                }
            ],
        },
    ]



from .bridge import *  # noqa
from .install import *  # noqa
from .list import *  # noqa
from .reflect import *  # noqa
from .run import *  # noqa
from .status import *  # noqa
from .clients import *  # noqa
from .connect import *  # noqa
from .operator import *  # noqa



import logging
from time import sleep
from typing import Dict, List, Optional

from docker.models.containers import Container

from gefyra.cli import console
from gefyra.configuration import ClientConfiguration
from gefyra.local.cargo import get_cargo_ip_from_netaddress
from gefyra.types import GefyraLocalContainer

from .utils import handle_docker_run_container

logger = logging.getLogger(__name__)


def handle_create_gefyrabridge(config: ClientConfiguration, body, target: str):
    from kubernetes.client import ApiException

    try:
        ireq = config.K8S_CUSTOM_OBJECT_API.create_namespaced_custom_object(
            namespace=config.NAMESPACE,
            body=body,
            group="gefyra.dev",
            plural="gefyrabridges",
            version="v1",
        )
    except ApiException as e:
        if e.status == 409:
            raise RuntimeError(f"Workload {target} already bridged.")
        logger.error(
            f"A Kubernetes API Error occured. \nReason: {e.reason} \nBody: {e.body}"
        )
        raise e from None
    return ireq


def handle_delete_gefyrabridge(config: ClientConfiguration, name: str) -> bool:
    from kubernetes.client import ApiException

    try:
        ireq = config.K8S_CUSTOM_OBJECT_API.delete_namespaced_custom_object(
            namespace=config.NAMESPACE,
            name=name,
            group="gefyra.dev",
            plural="gefyrabridges",
            version="v1",
        )
        return ireq
    except ApiException as e:
        if e.status == 404:
            logger.debug(f"InterceptRequest {name} not found")
        else:
            logger.debug("Error removing InterceptRequest: " + str(e))
        return False


def get_all_gefyrabridges(config: ClientConfiguration) -> list:
    from kubernetes.client import ApiException

    try:
        ireq_list = config.K8S_CUSTOM_OBJECT_API.list_namespaced_custom_object(
            namespace=config.NAMESPACE,
            group="gefyra.dev",
            plural="gefyrabridges",
            version="v1",
        )
        if ireq_list:
            # filter bridges for this client
            return list(
                item
                for item in ireq_list.get("items")
                if item["client"] == config.CLIENT_ID
            )
        else:
            return []
    except ApiException as e:
        if e.status != 404:
            logger.warning("Error getting GefyraBridges: " + str(e))
            raise e from None
        return []


def get_all_containers(config: ClientConfiguration) -> List[GefyraLocalContainer]:
    container_information = []
    gefyra_net = config.DOCKER.networks.get(f"{config.NETWORK_NAME}")
    containers = gefyra_net.containers
    # filter out gefyra-cargo container as well as fields other than name and ip
    for container in containers:
        if not container.name.startswith("gefyra-cargo"):
            container_information.append(
                GefyraLocalContainer(
                    name=container.name,
                    address=container.attrs["NetworkSettings"]["Networks"][
                        config.NETWORK_NAME
                    ]["IPAddress"].split("/")[0],
                    namespace=container.attrs["HostConfig"]["DnsSearch"][0].split(".")[
                        0
                    ],
                )
            )
    return container_information


def get_gbridge_body(
    config: ClientConfiguration,
    name: str,
    destination_ip,
    target_pod,
    target_namespace,
    target_container,
    port_mappings,
    handle_probes,
):
    return {
        "apiVersion": "gefyra.dev/v1",
        "kind": "gefyrabridge",
        "metadata": {
            "name": name,
            "namespace": config.NAMESPACE,
        },
        "provider": "carrier",
        "connectionProvider": "stowaway",
        "client": config.CLIENT_ID,
        "destinationIP": destination_ip,
        "targetPod": target_pod,
        "targetNamespace": target_namespace,
        "targetContainer": target_container,
        "portMappings": port_mappings,
        "handleProbes": handle_probes,
    }


def deploy_app_container(
    config: ClientConfiguration,
    image: str,
    name: str = "",
    command: str = "",
    volumes: Optional[List] = None,
    ports: Optional[Dict] = None,
    env: Optional[Dict] = None,
    auto_remove: bool = False,
    dns_search: Optional[List[str]] = None,
    pull: Optional[str] = "missing",
    platform: Optional[str] = "linux/amd64",
) -> Container:
    import docker

    if pull == "always":
        console.info(f"Pulling image {image}...")
        if ":" in image:
            repo, tag = image.split(":")
        else:
            repo, tag = image, "latest"
        config.DOCKER.images.pull(repository=repo, tag=tag, platform=platform)
        console.info(f"Pulling image {image} done.")

    if not dns_search:
        dns_search = ["default"]

    gefyra_net = config.DOCKER.networks.get(f"{config.NETWORK_NAME}")

    net_add = gefyra_net.attrs["IPAM"]["Config"][0]["Subnet"].split("/")[0]
    cargo_ip = get_cargo_ip_from_netaddress(net_add)
    all_kwargs = {
        "network": config.NETWORK_NAME,
        "name": name,
        "command": command,
        "volumes": volumes,
        "ports": ports,
        "detach": True,
        "dns": [config.STOWAWAY_IP],
        "dns_search": dns_search,
        "auto_remove": auto_remove,
        "environment": env,
        "pid_mode": f"container:{config.CARGO_CONTAINER_NAME}",  # noqa: E231
    }
    not_none_kwargs = {k: v for k, v in all_kwargs.items() if v is not None}

    container = handle_docker_run_container(config, image, **not_none_kwargs)

    cargo = config.DOCKER.containers.get(config.CARGO_CONTAINER_NAME)

    # busy wait for the container to start
    try:
        _i = 0
        while container.status == "created" and _i < (
            config.CONTAINER_RUN_TIMEOUT * 10
        ):
            sleep(0.1)
            container = config.DOCKER.containers.get(container.id)
            _i = _i + 1
    except docker.errors.NotFound:
        raise RuntimeError(
            f"Container {container.id} is not running. Did you miss a valid startup"
            " command?"
        )

    if container.status != "running":
        raise RuntimeError(
            "Container is not running. Did you miss a valid startup command?"
        )

    exit_code, output = cargo.exec_run(
        f"bash patchContainerGateway.sh {container.name} {cargo_ip}"
    )
    if exit_code == 0:
        logger.debug(f"Gateway patch applied to '{container.name}'")

    else:
        container = config.DOCKER.containers.get(container.id)
        if container.status != "running":
            raise RuntimeError(
                f"Container {name} is not running. Check whether your command is valid"
                " for the chosen image."
            )
        raise RuntimeError(
            f"Gateway patch could not be applied to '{container.name}': {output}"
        )
    return container



import dataclasses
import logging
from typing import Optional
import click

from gefyra.cli.utils import AliasedGroup, check_connection_name, standard_error_handler
from gefyra.cli import console
from tabulate import tabulate

logger = logging.getLogger(__name__)


def _manage_container_and_bridges(connection_name: str, force: bool = False):
    import kubernetes
    import urllib3
    from gefyra import api
    from gefyra.configuration import ClientConfiguration

    try:
        _bridges = api.list_gefyra_bridges(connection_name)
        if _bridges and len(_bridges[0][1]) > 0:
            console.info(
                f"There is {len(_bridges[0][1])} Gefyra bridge(s) running with connection '{connection_name}'."
            )
            if force:
                _del = True
            elif click.confirm("Do you want to remove them?", abort=True):
                _del = True
            if _del:
                for gbridges in _bridges[0][1]:
                    api.unbridge(
                        name=gbridges.name,
                        connection_name=connection_name,
                    )
    except (
        urllib3.exceptions.MaxRetryError,
        kubernetes.client.exceptions.ApiException,
    ):
        logger.warning("Cannot detect it there are any Gefyra bridges running")
    _containers = api.list_containers(connection_name)
    if _containers and len(_containers[0][1]) > 0:
        console.info(
            f"There is {len(_containers[0][1])} Gefyra container(s) running with connection '{connection_name}'."
        )
        if force:
            _del = True
        elif click.confirm("Do you want to remove them?", abort=True):
            _del = True
        if _del:
            for gcontainers in _containers[0][1]:
                container = ClientConfiguration(
                    connection_name=connection_name
                ).DOCKER.containers.get(gcontainers.name)
                container.remove(force=True)


@click.group(
    "connections",
    cls=AliasedGroup,
    help="Manage connections to clusters for this Gefyra installation",
)
@click.pass_context
def connections(ctx):
    pass


@connections.command(
    "connect",
    alias=["create"],
    help="Connect this local machine to a Gefyra cluster",
)
@click.option("-f", "--client-config", type=click.File("r"))
@click.option(
    "-n",
    "--connection-name",
    default="default",
    help="Assign a local name to this client connection",
    type=str,
)
@click.option(
    "--minikube",
    help="Connect Gefyra to a Minikube cluster (accepts minikube profile name, default is 'minikube'))",
    type=str,
    is_flag=False,
    flag_value="minikube",  # if --minikube is used as flag, we default to profile 'minikube'
    required=False,
)
@click.option(
    "--mtu",
    help="The MTU for the Wireguard interface",
    type=int,
    default=1340,
)
@standard_error_handler
def connect_client(
    client_config,
    connection_name: str,
    minikube: Optional[str] = None,
    mtu: int = 1340,
):
    from alive_progress import alive_bar
    from gefyra import api

    conn_list = api.list_connections()
    if (
        client_config
        and conn_list
        and connection_name in [conn.name for conn in conn_list]
    ):
        raise click.BadArgumentUsage(
            message=f"The connection name '{connection_name}' already exists. Run 'gefyra connections list' to "
            "see all connections. If this is a reconnect, please omit the --client-config option."
        )
    with alive_bar(
        total=None,
        length=20,
        title=f"Creating the cluster connection '{connection_name}'",
        bar="smooth",
        spinner="classic",
        stats=False,
        dual_line=True,
    ):
        api.connect(
            connection_name=connection_name,
            client_config=client_config,
            minikube_profile=minikube,
            mtu=mtu,
        )
    console.success(
        f"Connection established with connection name '{connection_name}'. "
        "Run 'gefyra connections list' to see all connections."
    )


@connections.command(
    "disconnect",
    alias=["stop", "halt"],
    help="Disconnect this local machine from a Gefyra cluster",
)
@click.argument(
    "connection_name", type=str, default="default", callback=check_connection_name
)
@standard_error_handler
def disconnect_client(connection_name: str):
    from gefyra import api

    _manage_container_and_bridges(connection_name=connection_name)
    api.disconnect(connection_name=connection_name)


@connections.command(
    "list",
    alias=["ls"],
    help="List all Gefyra connections",
)
@standard_error_handler
def list_connections():
    from gefyra import api

    conns = api.list_connections()
    data = [dataclasses.asdict(conn).values() for conn in conns]
    if data:
        click.echo(
            tabulate(
                data, headers=["NAME", "VERSION", "CREATED", "STATUS"], tablefmt="plain"
            )
        )
    else:
        console.info("No Gefyra connection found")


@connections.command(
    "remove",
    alias=["rm"],
    help="Remove a Gefyra connection",
)
@click.argument(
    "connection_name", type=str, default="default", callback=check_connection_name
)
# @standard_error_handler
def remove_connection(connection_name: str):
    from gefyra import api

    _manage_container_and_bridges(connection_name=connection_name)
    api.remove_connection(connection_name=connection_name)



from dataclasses import dataclass, field, fields
from enum import Enum
import json
import logging
from typing import Any, Dict, List, Optional

from gefyra.configuration import ClientConfiguration, __VERSION__
from gefyra.exceptions import ClientConfigurationError
from gefyra.local.clients import handle_get_gefyraclient

logger = logging.getLogger(__name__)


@dataclass
class StowawayParameter:
    # the subnet for a client
    subnet: str


@dataclass
class GefyraClientConfig:
    client_id: str
    kubernetes_server: str
    provider: str
    token: str
    namespace: str
    ca_crt: str
    gefyra_server: str
    registry: Optional[str] = None
    wireguard_mtu: Optional[str] = "1340"

    @property
    def json(self):
        return json.dumps(self.__dict__)

    @classmethod
    def from_json_str(cls, json_data: str):
        data = json.loads(json_data)
        return cls(**data)


@dataclass
class StowawayConfig:
    # the wireguard connection data
    # Interface.Address: 192.168.99.2
    iaddress: str
    # Interface.DNS: 192.168.99.1
    idns: str
    # Interface.ListenPort: 51820
    iport: int
    # Interface.PrivateKey: MFQ3v+...=
    iprivatekey: str
    # Peer.AllowedIPs: 0.0.0.0/0, ::/0
    pallowedips: str
    # Peer.Endpoint: 95.91.248.4:31820
    pendpoint: str
    # Peer.PublicKey: sy8jXi7...=
    ppublickey: str
    # Peer.PresharedKey: WCWY20...=
    presharedkey: str


class GefyraClientState(Enum):
    REQUESTED = "REQUESTED"
    CREATING = "CREATING"
    WAITING = "WAITING"
    ENABLING = "ENABLING"
    ACTIVE = "ACTIVE"
    DISABLING = "DISABLING"
    TERMINATING = "TERMINATING"
    ERROR = "ERROR"


@dataclass
class GefyraClient:
    # the id of the client
    client_id: str
    # the namespace this cluster runs in the host cluster
    namespace: str
    # the uid from Kubernetes for this object
    uid: str
    # the labels of this Gefyra object
    labels: Dict[str, str]
    # the provider of the client, always 'stowaway' in this version of Gefyra
    provider: str

    # the state of the client
    _state: str
    _state_transitions: Dict[str, str]
    provider_parameter: Optional[StowawayParameter] = None
    provider_config: Optional[StowawayConfig] = None
    service_account_name: Optional[str] = None
    service_account: Optional[Dict[str, str]] = None

    def __init__(self, gclient: dict[str, Any], config: ClientConfiguration):
        self._init_data(gclient)
        self._config = config

    def _init_data(self, _object: dict[str, Any]):
        self.client_id = _object["metadata"]["name"]
        self.uid = _object["metadata"]["uid"]
        self.provider = _object.get("provider", "")
        self._state = _object.get("state", "")
        self._state_transitions = _object.get("stateTransitions", {})
        self.service_account_name = _object.get("serviceAccountName")
        self.service_account = _object.get("serviceAccountData", {})
        if (
            providerparams := _object.get("providerParameter")
        ) and self.provider == "stowaway":
            self.provider_parameter = StowawayParameter(
                subnet=providerparams.get("subnet")
            )
        if (
            providerconfig := _object.get("providerConfig")
        ) and self.provider == "stowaway":
            self.provider_config = StowawayConfig(
                iaddress=providerconfig.get("Interface.Address"),
                idns=providerconfig.get("Interface.DNS"),
                iport=providerconfig.get("Interface.ListenPort"),
                iprivatekey=providerconfig.get("Interface.PrivateKey"),
                pallowedips=providerconfig.get("Peer.AllowedIPs"),
                pendpoint=providerconfig.get("Peer.Endpoint"),
                ppublickey=providerconfig.get("Peer.PublicKey"),
                presharedkey=providerconfig.get("Peer.PresharedKey"),
            )

    def as_dict(self) -> dict[str, Any]:
        data = {}
        for _field in fields(self):
            if _v := getattr(self, _field.name):
                if type(_v) is StowawayParameter:
                    data["providerParameter"] = {"subnet": _v.subnet}
                elif type(_v) is StowawayConfig:
                    data["providerConfig"] = {
                        "iaddress": _v.iaddress,
                        "idns": _v.idns,
                        "iport": str(_v.iport),
                        "iprivatekey": _v.iprivatekey,
                        "pallowedips": _v.pallowedips,
                        "pendpoint": _v.pendpoint,
                        "ppublickey": _v.ppublickey,
                    }
                else:
                    data[_field.name] = _v
        return data

    @property
    def state(self) -> GefyraClientState:
        self.update()
        return GefyraClientState(self._state)

    @property
    def state_transitions(self):
        self.update()
        return self._state_transitions

    def update(self):
        logger.debug(f"Fetching object GefyraClient {self.client_id}")
        gclient = handle_get_gefyraclient(self._config, self.client_id)
        self._init_data(gclient)

    def get_client_config(
        self,
        gefyra_server: str,
        k8s_server: str = "",
        registry: Optional[str] = None,
        wireguard_mtu: Optional[int] = 1340,
    ) -> GefyraClientConfig:
        if not bool(self.service_account):
            self.update()
        if self.service_account:
            return GefyraClientConfig(
                client_id=self.client_id,
                kubernetes_server=k8s_server or self._config.get_kubernetes_api_url(),
                provider=self.provider,
                token=self.service_account["token"],
                namespace=self.service_account["namespace"],
                ca_crt=self.service_account["ca.crt"],
                gefyra_server=gefyra_server,
                registry=registry,
                # if somehow the mtu is not given make sure to have null as json value
                wireguard_mtu=str(wireguard_mtu) if wireguard_mtu else None,
            )
        else:
            raise ClientConfigurationError(
                "Cannot get client config, no service account found."
            )

    def activate_connection(self, subnet: str):
        _state = self.state
        if _state == GefyraClientState.ACTIVE:
            return
        elif _state == GefyraClientState.WAITING:
            logger.debug(f"Activating connection for client {self.client_id}")
            self._config.K8S_CUSTOM_OBJECT_API.patch_namespaced_custom_object(
                group="gefyra.dev",
                version="v1",
                namespace=self._config.NAMESPACE,
                plural="gefyraclients",
                name=self.client_id,
                body={"providerParameter": {"subnet": subnet}},
            )
        else:
            raise RuntimeError(
                f"Cannot activate connection for client {self.client_id}, state is"
                f" {self.state}"
            )

    def deactivate_connection(self):
        _state = self.state
        if _state == GefyraClientState.WAITING:
            return
        elif _state == GefyraClientState.ACTIVE:
            logger.debug(f"Deactivating connection for client {self.client_id}")
            self._config.K8S_CUSTOM_OBJECT_API.patch_namespaced_custom_object(
                group="gefyra.dev",
                version="v1",
                namespace=self._config.NAMESPACE,
                plural="gefyraclients",
                name=self.client_id,
                body={"providerParameter": None},
            )
        else:
            raise RuntimeError(
                f"Cannot deactivate connection for client {self.client_id}, state is"
                f" {self.state}"
            )


@dataclass
class GefyraConnectionItem:
    name: str
    version: str
    created: str
    status: str


@dataclass
class GefyraInstallOptions:
    namespace: str = field(
        default_factory=lambda: "gefyra",
        metadata=dict(
            help="The namespace to install Gefyra into (default: gefyra)", short="ns"
        ),
    )
    version: str = field(
        default_factory=lambda: __VERSION__,
        metadata=dict(
            help=(
                "Set the Operator version; components are created according to this"
                f" Gefyra version (default: {__VERSION__})"
            )
        ),
    )
    service_type: str = field(
        default_factory=lambda: "NodePort",
        metadata=dict(
            help=(
                "The Kubernetes service for Stowaway to expose the Wireguard endpoint"
                " (default: NodePort)"
            )
        ),
    )
    service_port: int = field(
        default_factory=lambda: 31820,
        metadata=dict(
            help=(
                "The port for Stowaway to expose the Wireguard endpoint (default:"
                " 31820)"
            )
        ),
    )
    service_labels: Dict[str, str] = field(
        default_factory=lambda: {},
        metadata=dict(
            help="Additional Kubernetes labels for the Stowaway service (default: [])",
            type="array",
        ),
    )
    service_annotations: Dict[str, str] = field(
        default_factory=lambda: {},
        metadata=dict(
            help="Kubernetes annotations for the Stowaway service (default: [])",
            type="array",
        ),
    )
    registry: str = field(
        default_factory=lambda: "quay.io/gefyra",
        metadata=dict(
            help="The registry URL for the images (default: quay.io/gefyra)",
        ),
    )
    mtu: int = field(
        default_factory=lambda: 1340,
        metadata=dict(
            help="The MTU for the Wireguard interface (default: 1340)",
        ),
    )
    stowaway_storage: int = field(
        default_factory=lambda: 64,
        metadata=dict(
            help="The storage size for the Stowaway PVC in Mi (default: 64)",
        ),
    )


@dataclass
class GefyraClusterStatus:
    # is a kubernetes cluster reachable
    connected: bool
    # is the operator running
    operator: bool
    operator_image: str
    # is stowaway running
    stowaway: bool
    stowaway_image: str
    # the gefyra namespace is available
    namespace: bool
    # operator webhook
    operator_webhook: bool


@dataclass
class GefyraClientStatus:
    version: str
    # is cargo running
    cargo: bool
    cargo_image: str
    # is gefyra network available
    network: bool
    # is gefyra client connected with gefyra cluster
    connection: bool
    # amount of containers running in gefyra
    containers: int
    # amount of active bridges
    bridges: int
    # current kubeconfig file
    kubeconfig: str
    # current kubeconfig context
    context: str
    # wireguard endpoint
    cargo_endpoint: str


class StatusSummary(str, Enum):
    UP = "Gefyra is up and connected"
    DOWN = "Gefyra is not running"
    INCOMPLETE = "Gefyra is not running properly"


@dataclass
class GefyraStatus:
    summary: StatusSummary
    cluster: GefyraClusterStatus
    client: GefyraClientStatus


@dataclass
class GefyraLocalContainer:
    """
    A container managed(/started) by Gefyra
    """

    name: str
    address: str
    namespace: str


@dataclass
class GefyraBridge:
    """
    A GefyraBridge object
    """

    name: str
    client_id: str
    local_container_ip: str
    port_mappings: List[str]
    target_container: str
    target_namespace: str
    target_pod: str
    provider: str
    state: str



from pathlib import Path
from typing import Optional
from gefyra.api.utils import stopwatch
from gefyra.configuration import ClientConfiguration, __VERSION__
from gefyra.exceptions import ClusterError


@stopwatch
def update(
    version: Optional[str] = None,
    kubeconfig: Optional[Path] = None,
    kubecontext: Optional[str] = None,
) -> bool:
    # update operator deployment to latest version

    config = ClientConfiguration(kube_config_file=kubeconfig, kube_context=kubecontext)
    if not version:
        version = __VERSION__

    try:
        ns = config.K8S_CORE_API.read_namespaced_namespace(name=config.NAMESPACE)
        if ns.status.phase.lower() != "active":
            raise ClusterError(
                f"Cannot update Gefyra operator: namespace {config.NAMESPACE} is in {ns.status.phase} state"
            )
    except:  # noqa
        pass

    names = ["gefyra-operator", "gefyra-operator-webhook"]
    for name in names:
        config.K8S_APP_API.patch_namespaced_deployment(
            name=name,
            namespace=config.NAMESPACE,
            body={
                "spec": {
                    "template": {
                        "spec": {
                            "containers": [
                                {
                                    "name": "gefyra",
                                    "image": config.OPERATOR_IMAGE,
                                }
                            ]
                        }
                    }
                }
            },
        )
        config.K8S_APP_API.patch_namespaced_deployment(
            name=name,
            namespace=config.NAMESPACE,
            body={
                "spec": {
                    "template": {
                        "spec": {
                            "containers": [
                                {
                                    "name": "gefyra",
                                    "env": [
                                        {
                                            "name": "GEFYRA_STOWAWAY_TAG",
                                            "value": version,
                                        }
                                    ],
                                }
                            ]
                        }
                    }
                }
            },
        )
        config.K8S_APP_API.patch_namespaced_deployment(
            name=name,
            namespace=config.NAMESPACE,
            body={
                "spec": {
                    "template": {
                        "spec": {
                            "containers": [
                                {
                                    "name": "gefyra",
                                    "env": [
                                        {
                                            "name": "GEFYRA_CARRIER_IMAGE_TAG",
                                            "value": version,
                                        }
                                    ],
                                }
                            ]
                        }
                    }
                }
            },
        )

    return True



import logging
import click

from gefyra.misc.comps import COMPONENTS
from gefyra.api.install import LB_PRESETS

from gefyra.cli.utils import (
    installoptions_to_cli_options,
    multi_options,
    standard_error_handler,
)


@click.command(
    "install",
    help=(
        "Create and print the Kubernetes configs for Gefyra; usage: 'gefyra install"
        " [options] | kubectl apply -f -"
    ),
)
@click.option(
    "--component",
    "--comp",
    help=(
        "Limit config creation to this component (available:"
        f" {','.join([c.__name__.split('.')[-1] for c in COMPONENTS])})"
    ),
    type=str,
    multiple=True,
)
@click.option(
    "--preset",
    help=f"Set configs from a preset (available: {','.join(LB_PRESETS.keys())})",
    type=str,
)
@click.option(
    "--apply",
    is_flag=True,
)
@click.option(
    "--wait",
    is_flag=True,
)
@click.pass_context
@multi_options(installoptions_to_cli_options())
@standard_error_handler
def install(ctx, component, preset, apply, wait, **kwargs):
    from alive_progress import alive_bar
    from gefyra import api

    # filter out empty kwargs
    kwargs = {k: v for k, v in kwargs.items() if v}

    if wait and not apply:
        raise click.BadOptionUsage(
            option_name="wait", message="Cannot wait without '--apply'"
        )
    if wait:
        logger = logging.getLogger("gefyra")
        logger.setLevel(logging.INFO)
        formatter = logging.Formatter("%(message)s")
        if len(logger.handlers) > 0:
            logger.handlers[0].setFormatter(formatter)
        else:
            handler = logging.StreamHandler()
            handler.setFormatter(formatter)
            logger.addHandler(handler)
    if apply:
        with alive_bar(
            total=None,
            length=20,
            title="Installing Gefyra",
            bar="smooth",
            spinner="classic",
            stats=False,
            dual_line=True,
        ):
            ouput = api.install(component, preset, apply, wait, **kwargs)
    else:
        ouput = api.install(component, preset, apply, wait, **kwargs)
    if not apply:
        click.echo(ouput)


@click.command("uninstall", help="Removes the Gefyra installation from the cluster")
@click.option("--force", "-f", help="Delete without promt", is_flag=True)
@standard_error_handler
def uninstall(force):
    from alive_progress import alive_bar
    from gefyra import api

    if not force:
        click.confirm(
            "Do you want to remove all Gefyra components from this cluster?",
            abort=True,
        )
    with alive_bar(
        total=None,
        length=20,
        title="Uninstalling all Gefyra components from the cluster",
        bar="smooth",
        spinner="classic",
        stats=False,
        dual_line=True,
    ):
        api.uninstall()



import logging
from gefyra.exceptions import ClientConfigurationError
from gefyra.types import GefyraBridge, GefyraLocalContainer
from tabulate import tabulate
from typing import List, Optional, Tuple

from gefyra.configuration import ClientConfiguration

from .utils import stopwatch, wrap_bridge


logger = logging.getLogger(__name__)


def get_containers_and_print(connection_name: Optional[str] = None):
    from gefyra.api import list_containers

    containers = list_containers(connection_name=connection_name)
    print(
        tabulate(
            containers, headers=["NAME", "IP ADDRESS", "NAMESPACE"], tablefmt="plain"
        )
    )


def get_bridges_and_print(connection_name: Optional[str] = None):
    from gefyra.api import list_gefyra_bridges

    gefyra_bridges = list_gefyra_bridges(connection_name=connection_name)
    if gefyra_bridges:
        for gefyra_bridge in gefyra_bridges:
            print(gefyra_bridge)
    else:
        logger.info("No active bridges found")


@stopwatch
def list_gefyra_bridges(
    connection_name: Optional[str] = None,
) -> List[Tuple[str, List[GefyraBridge]]]:
    from gefyra.local.bridge import get_all_gefyrabridges
    from gefyra import api

    conns = api.list_connections()
    if connection_name:
        if connection_name not in [conns.name for conns in conns]:
            raise ClientConfigurationError(
                f"Connection {connection_name} does not exist. Please create it first."
            )
        obridges = get_all_gefyrabridges(
            ClientConfiguration(connection_name=connection_name)
        )
        return [(connection_name, list(map(wrap_bridge, obridges)))]
    else:
        bridges = []
        for conn in conns:
            bridges.append(
                (
                    conn.name,
                    list(
                        map(
                            wrap_bridge,
                            get_all_gefyrabridges(
                                ClientConfiguration(connection_name=conn.name)
                            ),
                        )
                    ),
                )
            )
        return bridges


@stopwatch
def list_containers(
    connection_name: Optional[str] = None,
) -> List[Tuple[str, List[GefyraLocalContainer]]]:
    from gefyra.local.bridge import get_all_containers
    from gefyra import api

    conns = api.list_connections()
    if connection_name:
        if connection_name not in [conns.name for conns in conns]:
            raise ClientConfigurationError(
                f"Connection {connection_name} does not exist. Please create it first."
            )
        return [
            (
                connection_name,
                get_all_containers(
                    ClientConfiguration(connection_name=connection_name)
                ),
            )
        ]
    else:
        return [
            (
                conn.name,
                get_all_containers(ClientConfiguration(connection_name=conn.name)),
            )
            for conn in conns
        ]



import dataclasses
from typing import Optional
import click

from gefyra import api
from gefyra.cli import console
from gefyra.cli.utils import standard_error_handler
from tabulate import tabulate


@click.command("status", help="List running containers and bridges")
@click.option(
    "-n",
    "--connection-name",
    help="Only list from this client connection name.",
    type=str,
)
@click.option(
    "-C",
    "--containers",
    help="Only list running containers.",
    type=bool,
    is_flag=True,
)
@click.option(
    "-B",
    "--bridges",
    help="Only list running bridges.",
    type=bool,
    is_flag=True,
)
@standard_error_handler
def list(
    connection_name: Optional[str] = None,
    containers: Optional[bool] = False,
    bridges: Optional[bool] = False,
):
    show_containers = True if containers is False and bridges is False else containers
    show_bridges = True if containers is False and bridges is False else bridges
    if show_containers:
        _containers = api.list_containers(connection_name)
        container_print = []
        for connection in _containers:
            for container in connection[1]:
                container_print.append(
                    {
                        "connection": connection[0],
                        **dataclasses.asdict(container),
                    }.values()
                )

        if container_print:
            if show_bridges:
                console.heading("Gefyra Containers:")
            click.echo(
                tabulate(
                    container_print,
                    headers=["CONNECTION", "CONTAINER NAME", "ADDRESS", "NAMESPACE"],
                    tablefmt="plain",
                )
            )
            if show_bridges:
                print()

        else:
            console.info("No local Gefyra containers found")

    if show_bridges:
        bridges_print = []
        _bridges = api.list_gefyra_bridges(connection_name)
        for connection in _bridges:
            for bridge in connection[1]:
                bridge_info = dataclasses.asdict(bridge)
                if bridge_info.get("port_mappings"):
                    bridge_info["port_mappings"] = "\n".join(
                        bridge_info["port_mappings"]
                    )
                bridges_print.append(
                    {"connection": connection[0], **bridge_info}.values()
                )

        if bridges_print:
            if show_containers:
                console.heading("Gefyra Bridges:")
            click.echo(
                tabulate(
                    bridges_print,
                    headers=[
                        "CONNECTION",
                        "NAME",
                        "CLIENT",
                        "LOCAL ADDRESS",
                        "PORTS",
                        "TARGET CONTAINER",
                        "NAMESPACE",
                        "TARGET POD",
                        "PROVIDER",
                        "STATE",
                    ],
                    tablefmt="plain",
                )
            )
        else:
            console.info("No Gefyra bridges found")



# flake8: noqa
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from gefyra.types import GefyraInstallOptions


def data(params: "GefyraInstallOptions") -> list[dict]:
    return [
        {
            "apiVersion": "v1",
            "kind": "Namespace",
            "metadata": {"name": params.namespace},
        }
    ]



from os import path
import os
import struct
import socket
import sys
import logging
from typing import Optional, Union

from pathlib import Path
from gefyra.exceptions import ClientConfigurationError


from gefyra.local import (
    CONNECTION_NAME_LABEL,
    CARGO_ENDPOINT_LABEL,
    ACTIVE_KUBECONFIG_LABEL,
    CLIENT_ID_LABEL,
)

logger = logging.getLogger("gefyra")

__VERSION__ = "2.3.0"
USER_HOME = os.path.expanduser("~")


def fix_pywin32_in_frozen_build() -> None:  # pragma: no cover
    import os
    import site

    if sys.platform != "win32" or not getattr(sys, "frozen", False):
        return

    site.addsitedir(sys.path[0])
    customsite = os.path.join(sys.path[0], "lib")
    site.addsitedir(customsite)

    # sys.path has been extended; use final
    # path to locate dll folder and add it to path
    path = sys.path[-1]
    path = path.replace("Pythonwin", "pywin32_system32")
    os.environ["PATH"] += ";" + path

    # import pythoncom module
    import importlib
    import importlib.machinery

    for name in ["pythoncom", "pywintypes"]:
        filename = os.path.join(path, name + "39.dll")
        loader = importlib.machinery.ExtensionFileLoader(name, filename)
        spec = importlib.machinery.ModuleSpec(name=name, loader=loader, origin=filename)
        importlib._bootstrap._load(spec)  # type: ignore


class ClientConfiguration(object):
    def __init__(
        self,
        docker_client=None,
        network_name: str = "",
        connection_name: Optional[str] = None,
        cargo_endpoint_host: str = "",
        cargo_endpoint_port: str = "31820",
        cargo_container_name: str = "",
        registry: Optional[str] = "",
        operator_image_url: str = "",
        stowaway_image_url: str = "",
        carrier_image_url: str = "",
        cargo_image_url: str = "",
        kube_config_file: Optional[Path] = None,
        kube_context: Optional[str] = None,
        wireguard_mtu: Optional[str] = None,
        client_id: str = "",
        gefyra_config_root: Optional[Union[str, Path]] = None,
        ignore_connection: bool = False,  # work with kubeconfig not connection
        ignore_docker: bool = False,
    ):
        import platform

        if sys.platform == "win32":  # pragma: no cover
            fix_pywin32_in_frozen_build()
        self.NAMESPACE = "gefyra"  # another namespace is currently not supported
        self._kube_config_path = None
        self._kube_context = None
        self.REGISTRY = registry.rstrip("/") if registry else "quay.io/gefyra"
        if registry:
            logger.debug(f"Using registry prefix (other than default): {self.REGISTRY}")
        self.OPERATOR_IMAGE = (
            operator_image_url or f"{self.REGISTRY}/operator:{__VERSION__}"
        )
        if operator_image_url:
            logger.debug(
                f"Using Operator image (other than default): {operator_image_url}"
            )
        self.STOWAWAY_IMAGE = (
            stowaway_image_url or f"{self.REGISTRY}/stowaway:{__VERSION__}"
        )
        if stowaway_image_url:
            logger.debug(
                f"Using Stowaway image (other than default): {stowaway_image_url}"
            )
        self.CARRIER_IMAGE = (
            carrier_image_url or f"{self.REGISTRY}/carrier:{__VERSION__}"
        )
        if carrier_image_url:
            logger.debug(
                f"Using Carrier image (other than default): {carrier_image_url}"
            )
        if sys.platform == "win32" or "microsoft" in platform.release().lower():
            self.CARGO_IMAGE = (
                cargo_image_url or f"{self.REGISTRY}/cargo-win:{__VERSION__}"
            )
        else:
            self.CARGO_IMAGE = cargo_image_url or f"{self.REGISTRY}/cargo:{__VERSION__}"
        if cargo_image_url:
            logger.debug(f"Using Cargo image (other than default): {cargo_image_url}")
        if docker_client:
            self.DOCKER = docker_client

        self.cargo_endpoint_port = cargo_endpoint_port

        self.CARGO_CONTAINER_NAME = cargo_container_name or "gefyra-cargo-default"
        self.STOWAWAY_IP = "192.168.99.1"
        self.NETWORK_NAME = network_name or "gefyra-network"
        self.CONNECTION_NAME = connection_name or "default"
        self.BRIDGE_TIMEOUT = 60  # in seconds
        self.CONNECTION_TIMEOUT = 60  # in seconds
        self.CARGO_PROBE_TIMEOUT = 20  # in seconds
        self.CONTAINER_RUN_TIMEOUT = 10  # in seconds
        self.CLIENT_ID = client_id
        if not ignore_docker:
            containers = self.DOCKER.containers.list(
                all=True,
                filters={"label": f"{CONNECTION_NAME_LABEL}={self.CONNECTION_NAME}"},
            )
            if containers and not ignore_connection:
                cargo_container = containers[0]
                self.CARGO_ENDPOINT = cargo_container.labels.get(CARGO_ENDPOINT_LABEL)
                self.KUBE_CONFIG_FILE = cargo_container.labels.get(
                    ACTIVE_KUBECONFIG_LABEL
                )
                self.CLIENT_ID = cargo_container.labels.get(CLIENT_ID_LABEL)
                self.CARGO_CONTAINER_NAME = cargo_container.name

        self.NETWORK_NAME = f"{self.NETWORK_NAME}-{self.CONNECTION_NAME}"
        if cargo_endpoint_host:
            self.CARGO_ENDPOINT = f"{cargo_endpoint_host}:{self.cargo_endpoint_port}"

        if kube_config_file:
            self.KUBE_CONFIG_FILE = str(kube_config_file)

        if kube_context:
            self.KUBE_CONTEXT = kube_context

        if not wireguard_mtu:
            self.WIREGUARD_MTU = "1340"
        else:
            self.WIREGUARD_MTU = wireguard_mtu
        if not gefyra_config_root:
            self.GEFYRA_LOCATION = Path.home().joinpath(".gefyra")
        else:
            self.GEFYRA_LOCATION = Path(gefyra_config_root)

    @property
    def CARGO_ENDPOINT(self):
        import platform

        if hasattr(self, "_cargo_endpoint") and self._cargo_endpoint:
            return self._cargo_endpoint
        else:
            if (
                platform.system().lower() == "linux"
                and "microsoft" not in platform.release().lower()
            ):
                # get linux docker0 network address
                import fcntl

                _soc = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                _ip = socket.inet_ntoa(
                    fcntl.ioctl(
                        _soc.fileno(),
                        0x8915,
                        struct.pack("256s", "docker0".encode("utf-8")[:15]),
                    )[20:24]
                )
                return f"{_ip}:{self.cargo_endpoint_port}"
            else:
                try:
                    _ip_output = self.DOCKER.containers.run(
                        "alpine", "getent hosts host.docker.internal", remove=True
                    )
                    _ip = _ip_output.decode("utf-8").split(" ")[0]
                    logger.debug(f"Found host.docker.internal IP: {_ip}")
                    return f"{_ip}:{self.cargo_endpoint_port}"
                except Exception as e:
                    logger.error("Could not create a valid configuration: " + str(e))

    @CARGO_ENDPOINT.setter
    def CARGO_ENDPOINT(self, value):
        self._cargo_endpoint = value

    @property
    def KUBE_CONTEXT(self):
        if not self._kube_context:
            from kubernetes.config.kube_config import list_kube_config_contexts
            from kubernetes.config.config_exception import ConfigException

            try:
                _, active_context = list_kube_config_contexts(
                    config_file=self.KUBE_CONFIG_FILE
                )
                self.KUBE_CONTEXT = active_context.get("name", None)
            except ConfigException:
                logger.error("Could not read active 'kubeconfig' context.")
                self.KUBE_CONTEXT = None
        return self._kube_context

    @KUBE_CONTEXT.setter
    def KUBE_CONTEXT(self, context):
        self._kube_context = context

    @property
    def KUBE_CONFIG_FILE(self):
        if not self._kube_config_path:
            from kubernetes.config.kube_config import KUBE_CONFIG_DEFAULT_LOCATION

            self.KUBE_CONFIG_FILE = KUBE_CONFIG_DEFAULT_LOCATION
        return self._kube_config_path

    @KUBE_CONFIG_FILE.setter
    def KUBE_CONFIG_FILE(self, kube_config_path):
        if not path.isfile(path.expanduser(kube_config_path)):
            raise RuntimeError(f"KUBE_CONFIG_FILE {kube_config_path} not found.")
        self._kube_config_path = kube_config_path

    def _init_docker(self):
        import docker
        from docker.context import ContextAPI

        try:
            ctx = ContextAPI.get_context()
            if ctx.name != "default":
                endpoint = ctx.endpoints["docker"]["Host"]
                self.DOCKER = docker.DockerClient(base_url=endpoint)
                logger.debug(f"Docker Context: {ctx.name}")
            else:
                self.DOCKER = docker.from_env()
        except docker.errors.DockerException as de:
            logger.fatal(f"Docker init error: {de}")
            raise RuntimeError("Docker init error. Docker host not running?") from None

    def _init_kubeapi(self):
        from kubernetes.client import (
            CoreV1Api,
            RbacAuthorizationV1Api,
            AppsV1Api,
            CustomObjectsApi,
            ApiextensionsV1Api,
            AdmissionregistrationV1Api,
        )
        from kubernetes.config import load_kube_config

        load_kube_config(self.KUBE_CONFIG_FILE, context=self.KUBE_CONTEXT)
        self.K8S_CORE_API = CoreV1Api()
        self.K8S_RBAC_API = RbacAuthorizationV1Api()
        self.K8S_APP_API = AppsV1Api()
        self.K8S_CUSTOM_OBJECT_API = CustomObjectsApi()
        self.K8S_EXTENSION_API = ApiextensionsV1Api()
        self.K8S_ADMISSION_API = AdmissionregistrationV1Api()

    def __getattr__(self, item):
        if item in [
            "K8S_CORE_API",
            "K8S_RBAC_API",
            "K8S_APP_API",
            "K8S_CUSTOM_OBJECT_API",
            "K8S_EXTENSION_API",
            "K8S_ADMISSION_API",
        ]:
            try:
                return self.__getattribute__(item)
            except AttributeError:
                self._init_kubeapi()
        if item == "DOCKER":
            try:
                return self.__getattribute__(item)
            except AttributeError:
                self._init_docker()

        return self.__getattribute__(item)

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if k.isupper()}

    def __str__(self):
        return str(self.to_dict())

    def get_kubernetes_api_url(self) -> str:
        return self.K8S_CORE_API.api_client.configuration.host

    def get_stowaway_host(self, port: Optional[str]) -> str:
        """
        Return the cargo endpoint
        If the endpoint is not set, it will try to estimate if from the K8s service
        gefyra-stowaway-wireguard in the cluster and/or public IPs of a node
        """
        if hasattr(self, "_cargo_endpoint") and self._cargo_endpoint:
            return self._cargo_endpoint
        else:
            import kubernetes

            try:
                service = self.K8S_CORE_API.read_namespaced_service(
                    namespace=self.NAMESPACE, name="gefyra-stowaway-wireguard"
                )
                if service.spec.type == "LoadBalancer":
                    _port = port or service.spec.ports["gefyra-wireguard"].port
                    _host = (
                        service.status.load_balancer.ingress[0].hostname
                        or service.status.load_balancer.ingress[0].ip
                    )
                    return f"{_host}:{_port}"
                else:  # NodePort
                    # trying to retrive a public IP for the service
                    nodes = self.K8S_CORE_API.list_node()
                    external_ips = list(
                        filter(
                            lambda x: x.type == "ExternalIP",
                            nodes.items[0].status.addresses,
                        )
                    )
                    if external_ips:
                        _port = port or service.spec.ports["gefyra-wireguard"].node_port
                        return f"{external_ips[0].address}:{_port}"
                    else:
                        raise ClientConfigurationError(
                            "Could not find a public IP for the NodePort service gefyra-stowaway-wireguard"
                        )
            except kubernetes.client.ApiException as e:
                if e.status == 404:
                    raise ClientConfigurationError(
                        f"Could not find service gefyra-stowaway-wireguard in {self.NAMESPACE}"
                    ) from None
                else:
                    raise e


def get_gefyra_config_location() -> str:
    """
    It creates a directory for the client config if it doesn't
    already exist, and returns the path to that directory

    :param config: ClientConfiguration
    :type config: ClientConfiguration
    :return: The path to the directory where the files will be stored.
    """
    config = ClientConfiguration()
    config_dir = config.GEFYRA_LOCATION
    config_dir.mkdir(parents=True, exist_ok=True)
    return str(config_dir)



import os
from typing import List, Optional, TYPE_CHECKING
import uuid

from gefyra.configuration import ClientConfiguration, logger
from gefyra.local import (
    BRIDGE_ID_LABEL,
    CLIENT_ID_LABEL,
    CONNECTION_NAME_LABEL,
    CREATED_BY_LABEL,
    ACTIVE_KUBECONFIG_LABEL,
    ACTIVE_KUBECONFIG_CONTEXT_LABEL,
    CARGO_ENDPOINT_LABEL,
    VERSION_LABEL,
    CARGO_LABEL,
)

if TYPE_CHECKING:
    from docker.models.containers import Container


def get_processed_paths(base_path: str, volumes: List[str]) -> Optional[List[str]]:
    if volumes is None:
        return None
    results = []
    for volume in volumes:
        source, target = volume.rsplit(":", 1)
        if not os.path.isabs(source):
            source = os.path.realpath(os.path.join(base_path, source))
        results.append(f"{source}:{target}")
    return results


def handle_docker_get_or_create_container(
    config: ClientConfiguration, name: str, image: str, **kwargs
) -> "Container":
    import docker

    try:
        return config.DOCKER.containers.get(name)
    except docker.errors.NotFound:
        return handle_docker_create_container(config, image, name=name, **kwargs)


def handle_docker_create_container(
    config: ClientConfiguration, image: str, **kwargs
) -> "Container":
    import gefyra.configuration
    import docker

    try:
        config.DOCKER.images.get(image)
    except docker.errors.ImageNotFound:
        repo, tag = image.split(":")
        logger.debug(f"Pulling {repo}:{tag} image.")
        config.DOCKER.images.pull(repository=repo, tag=tag)

    return config.DOCKER.containers.create(
        image,
        labels={
            CREATED_BY_LABEL[0]: CREATED_BY_LABEL[1],
            CARGO_LABEL[0]: CARGO_LABEL[1],
            ACTIVE_KUBECONFIG_LABEL: config.KUBE_CONFIG_FILE,
            ACTIVE_KUBECONFIG_CONTEXT_LABEL: config.KUBE_CONTEXT,
            CARGO_ENDPOINT_LABEL: config.CARGO_ENDPOINT,
            CONNECTION_NAME_LABEL: config.CONNECTION_NAME,
            CLIENT_ID_LABEL: config.CLIENT_ID,
            VERSION_LABEL: gefyra.configuration.__VERSION__,
        },
        **kwargs,
    )


def handle_docker_run_container(
    config: ClientConfiguration, image: str, **kwargs
) -> "Container":
    # if detach=True is in kwargs, this will return a container; otherwise the container logs (see
    # https://docker-py.readthedocs.io/en/stable/containers.html#docker.models.containers.ContainerCollection.run)
    # TODO: handle exception(s):
    # docker.errors.ContainerError – If the container exits with a non-zero exit code and detach is False.
    # docker.errors.ImageNotFound – If the specified image does not exist.
    # docker.errors.APIError – If the server returns an error.
    return config.DOCKER.containers.run(
        image,
        labels={
            CREATED_BY_LABEL[0]: CREATED_BY_LABEL[1],
            BRIDGE_ID_LABEL: uuid.uuid4().hex,
        },
        **kwargs,
    )


def get_connection_from_kubeconfig(kubeconfig: Optional[str] = None) -> Optional[str]:
    import yaml

    if kubeconfig:
        _file = kubeconfig
    else:
        from kubernetes.config.kube_config import KUBE_CONFIG_DEFAULT_LOCATION
        from pathlib import Path

        _file = str(Path(KUBE_CONFIG_DEFAULT_LOCATION).expanduser())

    try:
        with open(_file, "r") as kubeconfig_file:
            kubecfg = yaml.safe_load(kubeconfig_file)
        active_ctx = next(
            filter(
                lambda x: x["name"] == kubecfg["current-context"], kubecfg["contexts"]
            )
        )
        if gefyra_connection := active_ctx.get("gefyra"):
            return gefyra_connection
        else:
            return None
    except Exception as e:  # noqa
        logger.error(
            f"Could not load Gefyra --host and --port from kubeconfig due to: {e}"
        )
        return None


def compose_kubeconfig_for_serviceaccount(
    server: str, ca: str, namespace: str, token: str
):
    return f"""apiVersion: v1
kind: Config
clusters:
  - name: default-cluster
    cluster:
      certificate-authority-data: {ca}
      server: {server}
contexts:
  - name: default-context
    context:
      cluster: default-cluster
      namespace: {namespace}
      user: default-user
current-context: default-context
users:
  - name: default-user
    user:
      token: {token}
    """



import click


@click.group(
    "self",
    help="Manage local Gefyra client installation",
)
@click.pass_context
def _self(ctx: click.Context):
    pass


@_self.command(
    "restore",
    help="Restore the local Gefyra client installation",
)
def restore():
    pass


@_self.command(
    "update",
    help="Update the local Gefyra client installation",
)
def update():
    pass
